"""
Adapted from NiftyNet
"""

import numpy as np
import torch
from ..torchio import INTENSITY
from ..utils import is_image_dict
from .random_transform import RandomTransform


class RandomBiasField(RandomTransform):
    def __init__(
            self,
            coefficients_range=(-0.5, 0.5),
            order=3,
            seed=None,
            verbose=False,
            ):
        super().__init__(seed=seed, verbose=verbose)
        self.coefficients_range = coefficients_range
        self.order = order

    def apply_transform(self, sample):
        coefficients = self.get_params(self.order, self.coefficients_range)
        sample['random_bias_field'] = coefficients
        for image_name, image_dict in sample.items():
            if not is_image_dict(image_dict):
                continue
            if image_dict['type'] != INTENSITY:
                continue
            coefficients = self.get_params(self.order, self.coefficients_range)
            sample[image_name]['random_bias_field'] = coefficients
            bias_field = self.generate_bias_field_map(
                image_dict['data'], self.order, coefficients)
            image_with_bias = image_dict['data'] * torch.from_numpy(bias_field)
            image_dict['data'] = image_with_bias
        return sample

    @staticmethod
    def get_params(order, coefficients_range):
        """
        Sampling of the appropriate number of coefficients for the creation
        of the bias field map
        """
        random_coefficients = []
        for x_order in range(0, order + 1):
            for y_order in range(0, order + 1 - x_order):
                for z_order in range(0, order + 1 - (x_order + y_order)):
                    number = torch.FloatTensor(1).uniform_(*coefficients_range)
                    random_coefficients.append(number.item())
        return np.array(random_coefficients)

    @staticmethod
    def generate_bias_field_map(data, order, coefficients):
        """
        Create the bias field map using a linear combination of polynomial
        functions and the coefficients previously sampled
        """
        shape = np.array(data.shape[1:])  # first axis is channels
        half_shape = shape / 2

        ranges = [np.arange(-n, n) for n in half_shape]

        bf_map = np.zeros(shape)
        x_mesh, y_mesh, z_mesh = np.asarray(np.meshgrid(*ranges))

        x_mesh /= x_mesh.max()
        y_mesh /= y_mesh.max()
        z_mesh /= z_mesh.max()

        i = 0
        for x_order in range(order + 1):
            for y_order in range(order + 1 - x_order):
                for z_order in range(order + 1 - (x_order + y_order)):
                    random_coefficient = coefficients[i]
                    new_map = (
                        random_coefficient
                        * x_mesh ** x_order
                        * y_mesh ** y_order
                        * z_mesh ** z_order
                    )
                    bf_map += np.transpose(new_map, (1, 0, 2))  # why?
                    i += 1
        return np.exp(bf_map)



"""Top-level package for torchio."""

__author__ = """Fernando Perez-Garcia"""
__email__ = 'fernando.perezgarcia.17@ucl.ac.uk'
__version__ = '0.7.8'

from . import io
from . import utils
from . import sampler
from . import inference
from .torchio import *
from .queue import Queue
from .transforms import *
from .dataset import ImagesDataset



#!/usr/bin/env python

"""Tests for ImagesDataset."""

import random
import tempfile
import unittest
from pathlib import Path
import numpy as np
import nibabel as nib
import torchio
from torchio import INTENSITY, LABEL


class TestRandomElasticDeformation(unittest.TestCase):
    """Tests for `RandomElasticDeformation`."""

    def setUp(self):
        """Set up test fixtures, if any."""
        self.dir = Path(tempfile.gettempdir()) / '.torchio_tests'
        self.dir.mkdir(exist_ok=True)
        random.seed(42)
        np.random.seed(42)

        subject_a = {
            't1': dict(path=self.get_image_path('t1_a'), type=INTENSITY),
        }
        subject_b = {
            't1': dict(path=self.get_image_path('t1_b'), type=INTENSITY),
            'label': dict(path=self.get_image_path('label_b'), type=LABEL),
        }
        subject_c = {
            'label': dict(path=self.get_image_path('label_c'), type=LABEL),
        }
        subject_d = {
            't1': dict(path=self.get_image_path('t1_d'), type=INTENSITY),
            't2': dict(path=self.get_image_path('t2_d'), type=INTENSITY),
            'label': dict(path=self.get_image_path('label_d'), type=LABEL),
        }
        self.paths_list = [
            subject_a,
            subject_b,
            subject_c,
            subject_d,
        ]

    def tearDown(self):
        """Tear down test fixtures, if any."""
        import shutil
        shutil.rmtree(self.dir)

    def get_image_path(self, stem):
        data = np.random.rand(10, 20, 30)
        affine = np.eye(4)
        suffix = random.choice(('.nii.gz', '.nii'))
        path = self.dir / f'{stem}{suffix}'
        nib.Nifti1Image(data, affine).to_filename(str(path))
        path = str(path) if np.random.rand() > 0.5 else path
        return path

    def test_images(self):
        self.iterate_dataset(self.paths_list)

    def test_wrong_paths_list(self):
        with self.assertRaises(ValueError):
            self.iterate_dataset([])
        with self.assertRaises(ValueError):
            self.iterate_dataset(())
        with self.assertRaises(TypeError):
            self.iterate_dataset(0)
        with self.assertRaises(TypeError):
            self.iterate_dataset([0])
        with self.assertRaises(ValueError):
            self.iterate_dataset([{}])
        with self.assertRaises(ValueError):
            self.iterate_dataset([{}, {}])
        with self.assertRaises(FileNotFoundError):
            self.iterate_dataset([{'t1':dict(path='nopath', type=INTENSITY)}])
        with self.assertRaises(TypeError):
            self.iterate_dataset([{'t1': dict(path=5, type=INTENSITY)}])
        with self.assertRaises(ValueError):
            path = self.dir / 'test.txt'
            path.touch()
            self.iterate_dataset([{'t1': dict(path=path, type=INTENSITY)}])
        with self.assertRaises(KeyError):
            self.iterate_dataset([{'t1': dict(path='nopath')}])
        with self.assertRaises(KeyError):
            self.iterate_dataset([{'t1': dict(type=INTENSITY)}])
        with self.assertRaises(KeyError):
            self.iterate_dataset([{'t1': dict(test='', type=INTENSITY)}])
        with self.assertRaises(TypeError):
            self.iterate_dataset([{'t1': 6}])

    def test_others(self):
        dataset = torchio.ImagesDataset(
            self.paths_list, verbose=True, transform=lambda x: x)
        _ = len(dataset)  # for coverage
        sample = dataset[0]
        output_path = self.dir / 'test.nii.gz'
        paths_dict = {'t1': output_path}
        dataset.save_sample(sample, paths_dict)
        nii = nib.load(str(output_path))
        ndims_output = len(nii.shape)
        ndims_sample = len(sample['t1']['data'].shape)
        assert ndims_sample == ndims_output + 1

    def iterate_dataset(self, paths_list):
        dataset = torchio.ImagesDataset(paths_list)
        for _ in dataset:
            pass
